
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosClock.pas                                         }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosClock.pas,v 1.2 2003/10/25 22:00:07 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosClock.pas,v $
  Revision 1.2  2003/10/25 22:00:07  iwache
  Minor changes in coding format added. No functional change.

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
{
  File Description : This unit provides basic clock-handling
  and computation of time differences. It also allows the
  programmer to change the timer's precision. The default
  precision is about 31 milliseconds, although it ranges
  from 500 ms to 100 microseconds.
}
unit WDosClock;

interface

uses
  WDosPorts, WDosInterrupts, SysUtils;

type
  TTimerPrecision =
    (tp2, tp4, tp8, tp16, tp32, tp64, tp128, tp256, tp512, tp1024,
    tp2048, tp4096, tp8192);

function GetMSCounter: Int64;
procedure SetTimerPrecision(Precision: TTimerPrecision);

implementation

var
  _Divider: Integer;
  _MSCounter: Int64;

procedure WriteClock(Address, Value: Byte); forward;

{******************************** Function Header ******************************
Function Name: GetMSCounter
Author       : Diego De Marco <ddemarco@dpsautomation.com>
Date Created : 30/01/2003
Description  : Returns the current status of the clock counter, in milliseconds.
Parameters   : None
Return Value : Clock count, in milliseconds. This number represents the time
      ellapsed since the program was started. It is not affected by
      the RTC time. In other words, this function is not affected by
      changes in the system date and time.
*******************************************************************************}

function GetMSCounter: Int64;
begin
  Result := _MSCounter * 1000;
  Result := Result div _Divider;
end;

{******************************** Function Header ******************************
Function Name: SetTimerPrecision
Author       : Diego De Marco <ddemarco@dpsautomation.com>
Date Created : 30/01/2003
Description  : Changes the precision of the internal timer.
Parameters   : Precision: New precision to use, from 500 ms to 100 us. The
      parameter represents a divider. For instance, instance, issuing
      SetTimerPrecision(tp512) means 1.95 milliseconds (1000/512).
Return Value : None
*******************************************************************************}

procedure SetTimerPrecision(Precision: TTimerPrecision);
var
  OldDivider: Integer;
begin
  OldDivider := _Divider;

  { Change Clock Speed and calculate new divider }
  { Register A of the RTC controls the speed }
  {
  Bit 7
   Indicates that clock is being updated (Time should not be read now)

  Bits 6-4
    --------
   010     use 32,768 Hz xtal
   others  don't do anything useful in PCs, really...

    Bits 3-0
    --------
   0000    seems to inhibit the whole RTC
   0001    divides xtal by 128 (PC: 256 ints per second)
   0010    divides xtal by 256 (PC: 128 ints per second)
   0011    divides xtal by 4 (PC: 8192 ints per second)
   0100    divides xtal by 8 (PC: 4096 ints per second)
   0101    divides xtal by 16 (PC: 2048 ints per second)
   0110    divides xtal by 32 (PC: 1024 ints per second)
   0111    divides xtal by 64 (PC: 512 ints per second)
   1000    divides xtal by 128 (PC: 256 ints per second)
   1001    divides xtal by 256 (PC: 128 ints per second)
   1010    divides xtal by 512 (PC: 64 ints per second)
   1011    divides xtal by 1024 (PC: 32 ints per second)
   1100    divides xtal by 2048 (PC: 16 ints per second)
   1101    divides xtal by 4096 (PC: 8 ints per second)
   1110    divides xtal by 8192 (PC: 4 ints per second)
   1111    divides xtal by 16384 (PC: 2 ints per second)
  }
  WriteClock($0A, $20 { Means "use 32 kHz clock" } or $0D
    { Means "divide clock by 1024" });
  case Precision of
    tp2:
      begin
        WriteClock($0A, $20 or $0F);
        _Divider := 2;
      end;
    tp4:
      begin
        WriteClock($0A, $20 or $0E);
        _Divider := 4;
      end;
    tp8:
      begin
        WriteClock($0A, $20 or $0D);
        _Divider := 8;
      end;
    tp16:
      begin
        WriteClock($0A, $20 or $0C);
        _Divider := 16;
      end;
    tp32:
      begin
        WriteClock($0A, $20 or $0B);
        _Divider := 32;
      end;
    tp64:
      begin
        WriteClock($0A, $20 or $0A);
        _Divider := 64;
      end;
    tp128:
      begin
        WriteClock($0A, $20 or $09);
        _Divider := 128;
      end;
    tp256:
      begin
        WriteClock($0A, $20 or $08);
        _Divider := 256;
      end;
    tp512:
      begin
        WriteClock($0A, $20 or $07);
        _Divider := 512;
      end;
    tp1024:
      begin
        WriteClock($0A, $20 or $06);
        _Divider := 1024;
      end;
    tp2048:
      begin
        WriteClock($0A, $20 or $05);
        _Divider := 2048;
      end;
    tp4096:
      begin
        WriteClock($0A, $20 or $04);
        _Divider := 4096;
      end;
    tp8192:
      begin
        WriteClock($0A, $20 or $03);
        _Divider := 8192;
      end;
  end;

  { Recalculate _MSCounter }
  if _MScounter <> 0 then
    _MSCounter := _MSCounter * _Divider div OldDivider;
end;

{**************************************************************************}
{							Internal Functions							   }
{**************************************************************************}
var
  OldClockHandler: DpmiPmVector;
  RegAValue, RegBValue, RegA1Value: Byte;

function ReadClock(Address: Byte): Byte;
begin
  Port[$70] := Address;
  ReadClock := Port[$71];
end;

procedure WriteClock(Address, Value: Byte);
begin
  Port[$70] := Address;
  Port[$71] := Value;
end;

procedure ClearClock;
begin
  ReadClock($0C);
end;

procedure ClockHandler;
begin
  Inc(_MSCounter);
  ClearClock;
  // Enable interrupts
  Port[$20] := $20;
  Port[$A0] := $20;
end;

procedure Clock_Install;
begin
  { Save Clock Parameters }
  RegAValue := ReadClock($0A);
  RegBValue := ReadClock($0B);
  RegA1Value := Port[$A1];
  WriteClock($0B, RegBValue and $BF); { Disable clock interrupt }

  { Initialize variables }
  _MSCounter := 0;

  { Install new handler }
  GetIntVec($70, OldClockHandler);
  SetIntProc($70, ClockHandler);

  { Set clock speed }
  SetTimerPrecision(tp32);

  { Enable clock interrupts }
  Port[$A1] := Port[$A1] and $FE;
  WriteClock($0B, RegBValue or 64);
  ClearClock;
end;

procedure Clock_Uninstall;
begin
  { Restore clock rate }
  WriteClock($0A, RegAValue);
  WriteClock($0B, RegBValue);
  Port[$A1] := RegA1Value;

  { Restore interrupts }
  SetIntVec($70, OldClockHandler);
end;

initialization
  Clock_Install;

finalization
  Clock_Uninstall;
end.

