
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosResolvers.pas                                     }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{                                                       }
{ Contributor(s):                                       }
{   Diego De Marco ddemarco@dpsautomation.com           }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosResolvers.pas,v 1.1 2003/06/11 21:20:39 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosResolvers.pas,v $
  Revision 1.1  2003/06/11 21:20:39  iwache
  A class Resolver for hosts and services resolution.
  Hosts works together with DNS name resolving. First
  it looks for an entry in file hosts then ask the DNS server.
  The files hosts, services and network.ini are searched in
  following order:
  1. in the current directory
  2. in the /etc folder of the current drive
  3. in c:/etc folder.

  ----------------------------------------------------------------------------
}
unit WDosResolvers;

interface

uses
  SysUtils, Classes, Contnrs, WDosSock, WDosIP;

const
  SWorkingDir = '.\';
  SDriveDir = '\etc\';
  SDriveCDir = 'c:\etc\';

  SHostFileName = 'hosts';
  SServicesFileName = 'services';
  SNetworkFileName = 'network.ini';

type
  TSocketEntry = class(TObject)
  private
    FHostName: string;
    FServiceName: string;
    FIpAddr: TInAddr;
    FPort: Word;
  public
    property HostName: string read FHostName;
    property IpAddr: TInAddr read FIpAddr;
    property ServiceName: string read FServiceName;
    property Port: Word read FPort;
  end;

  TLookupResult = (clrOk, clrNetDown, clrInval, clrAlready);

  TResolver = class(TObject)
  private
    FHosts: TStringList;
    FServs: TObjectList;
    FHostName: string;
    FDhcp: Boolean;
    FIpAddress: TIPAddress;
    FNetMask: TIPAddress;
    FGateway: TIPAddress;
    FNameServer: TIPAddress;
    FDomainName: string;
    function FindFileName(const FileName: string;
        var FullPath: string): Boolean;
    procedure LoadLinesFromFile(Lines: TStrings; const HostFileName: string);
    procedure LoadHostsFromFile;
    procedure LoadServicesFromFile;
    procedure LoadNetworkFromFile;
    procedure AddHost(const AHostName: string; AIpAddr: TInAddr);
    procedure AddPort(const AServName: string; APort: Word;
      const AProtocol: string);
  protected
  public
    constructor Create;
    destructor Destroy; override;
    function GetHostByName(const AName: string): TInAddr;
    function GetHostByAddr(Addr: TInAddr): string;
    function GetServByName(const AName, AProtocol: string): Word;
    function GetServByPort(APort: Word; const AProtocol: string): string;
    function AppendDomain(const HostName: string): string;
    property HostName: string read FHostName;
    property Dhcp: Boolean read FDhcp;
    property IpAddress: TIPAddress read FIpAddress;
    property NetMask: TIPAddress read FNetMask;
    property Gateway: TIPAddress read FGateway;
    property NameServer: TIPAddress read FNameServer;
    property DomainName: string read FDomainName;
  end;

var
  Resolver: TResolver;

implementation

uses
  IniFiles, WDosDNS;

const
  Delimiters = [#09, ' ', ',', ';'];

type
  TCharSet = set of Char;

  TService = class(TObject)
    Name: string;
    Port: Word;
    Protocol: string;
  end;

function StripRemarks(const Text: string; RemarkToken: Char): string;
var
  RemarkPos: Integer;
begin
  RemarkPos := Pos(RemarkToken, Text);
  if RemarkPos = 0 then
    Result := Text
  else
    Result := Copy(Text, 1, RemarkPos - 1);
end;

function Tokenize(Strings: TStrings; const Line: string;
  Delimiters: TCharSet): Integer;
var
  P, Start: PChar;
  S: string;
begin
  Result := 0;
  if not Assigned(Strings) then
    Exit;
  with Strings do
  begin
    Clear;
    P := Pointer(Line);
    if P <> nil then
      while P^ <> #0 do
      begin
        Start := P;
        while not (P^ in (Delimiters + [#0])) do
          Inc(P);
        SetString(S, Start, P - Start);
        Add(S);
        while P^ in Delimiters do
          Inc(P);
      end;
    Result := Count;
  end;
end;

function SplitToken(const Text: string; SplitChar: Char;
  var First, Last: string): Boolean;
var
  SplitPos: Integer;
begin
  Result := False;
  SplitPos := Pos(SplitChar, Text);
  if SplitPos = 0 then
    Exit;
  First := Trim(Copy(Text, 1, SplitPos - 1));
  Last := Trim(Copy(Text, SplitPos + 1, Length(Text) - SplitPos));
  Result := (First <> '') and (Last <> '');
end;

{ TResolver }

constructor TResolver.Create;
begin
  inherited Create;
  FHosts := TStringList.Create;
  FServs := TObjectList.Create;
  FDhcp := False;
  FIpAddress := INADDR_ANY;
  FNetMask := IP_StringToAddr('255.255.255.0');
  FGateway := INADDR_ANY;
  FHostName := 'localhost';
  FDomainName := 'localdomain';
  FNameServer := INADDR_ANY;

  LoadHostsFromFile;
  LoadServicesFromFile;
  LoadNetworkFromFile;
end;

destructor TResolver.Destroy;
begin
  FServs.Free;
  FHosts.Free;
  inherited Destroy;
end;

procedure TResolver.AddHost(const AHostName: string; AIpAddr: TInAddr);
begin
  FHosts.AddObject(AHostName, Pointer(AIpAddr));
end;

procedure TResolver.AddPort(const AServName: string; APort: Word;
  const AProtocol: string);
var
  Service: TService;
begin
  Service := TService.Create;
  with Service do
  begin
    Name := AServName;
    Port := APort;
    Protocol := AProtocol;
  end;
  FServs.Add(Service);
end;

function TResolver.GetHostByAddr(Addr: TInAddr): string;
var
  Index: Integer;
begin
  Index := FHosts.IndexOfObject(TObject(Addr));
  if Index < 0 then
    Result := ''
  else
    Result := FHosts[Index];
end;

function TResolver.GetHostByName(const AName: string): TInAddr;
var
  Index: Integer;
  Dest: TIPAddress;
begin
  // first search in hosts file
  Index := FHosts.IndexOf(AName);
  if Index < 0 then
    Result.S_addr := INADDR_NONE
  else
    Result := TInAddr(FHosts.Objects[Index]);

  // second use NAMESERVER if available
  if (Result.S_addr = INADDR_NONE) and (FNameServer <> INADDR_ANY) then
  begin
    if WDosDNS.GetHostByName(AName, Dest) then
    begin
      Result.S_addr := Dest;
      Exit;
    end;
    if WDosDNS.GetHostByName(AppendDomain(AName), Dest) then
      Result.S_addr := Dest;
  end;
end;

function TResolver.GetServByName(const AName, AProtocol: string): Word;
var
  I: Integer;
begin
  for I := 0 to FServs.Count - 1 do
    with TService(FServs[I]) do
    begin
      Result := Port;
      if (AnsiCompareText(AName, Name) = 0)
          and ((AProtocol = '')
          or (AnsiCompareText(AProtocol, Protocol) = 0)) then
        Exit;
    end;
  Result := 0;
end;

function TResolver.GetServByPort(APort: Word; const AProtocol: string): string;
var
  I: Integer;
begin
  for I := 0 to FServs.Count - 1 do
    with TService(FServs[I]) do
    begin
      Result := Name;
      if (APort = Port)
        and (AnsiCompareText(AProtocol, Protocol) = 0) then
        Exit;
    end;
  Result := '';
end;

function TResolver.FindFileName(const FileName: string;
    var FullPath: string): Boolean;
begin
  FullPath := SWorkingDir + FileName;
  Result := FileExists(FullPath);
  if Result then Exit;

  FullPath := SDriveDir + FileName;
  Result := FileExists(FullPath);
  if Result then Exit;

  FullPath := SDriveCDir + FileName;
  Result := FileExists(FullPath);
end;

procedure TResolver.LoadLinesFromFile(Lines: TStrings;
  const HostFileName: string);
var
  FullPath: string;
begin
  try
    if FindFileName(HostFileName, FullPath) then
      Lines.LoadFromFile(FullPath);
  except
    on E: Exception do
      Writeln(E.Message);
  end;
end;

procedure TResolver.LoadHostsFromFile;
var
  I, K: Integer;
  S: string;
  Lines, Tokens: TStringList;
  NumTokens: Integer;
  IpAddr: TInAddr;
begin
  Lines := TStringList.Create;
  try
    // load hosts file
    LoadLinesFromFile(Lines, SHostFileName);
    // remove remarks and rem lines
    for I := Lines.Count - 1 downto 0 do
    begin
      S := StripRemarks(Lines[I], '#');
      if S = '' then
        Lines.Delete(I)
      else
        Lines[I] := Trim(S);
    end;
    // tokenize every line and store IP and Host
    Tokens := TStringList.Create;
    try
      for I := 0 to Lines.Count - 1 do
      begin
        // resolve string into tokens
        NumTokens := Tokenize(Tokens, Lines[I], Delimiters);
        // format tokens
        if NumTokens > 1 then
        begin
          IpAddr.S_addr := inet_addr(PChar(Tokens[0]));
          for K := 1 to NumTokens - 1 do
            AddHost(Tokens[K], IpAddr);
        end;
      end;
    finally
      Tokens.Free;
    end;
  finally
    Lines.Free;
  end;
end;

procedure TResolver.LoadServicesFromFile;
var
  I, K: Integer;
  S: string;
  Lines, Tokens: TStringList;
  NumTokens: Integer;
  Name: string;
  Port: Word;
  Protocol: string;
begin
  Lines := TStringList.Create;
  try
    // load services file
    LoadLinesFromFile(Lines, SServicesFileName);
    // remove remarks and rem lines
    for I := Lines.Count - 1 downto 0 do
    begin
      S := StripRemarks(Lines[I], '#');
      if S = '' then
        Lines.Delete(I)
      else
        Lines[I] := Trim(S);
    end;
    // tokenize every line and store Name, Port and Protocol
    Tokens := TStringList.Create;
    try
      for I := 0 to Lines.Count - 1 do
      begin
        // String in Token auflsen
        NumTokens := Tokenize(Tokens, Lines[I], Delimiters);
        // Token formatieren
        if NumTokens > 1 then
        begin
          Name := Tokens[0];
          if not SplitToken(Tokens[1], '/', S, Protocol) then
            Continue;
          Port := htons(StrToIntDef(S, 0));
          AddPort(Name, Port, Protocol);
          for K := 2 to NumTokens - 1 do
            AddPort(Tokens[K], Port, Protocol);
        end;
      end;
    finally
      Tokens.Free;
    end;
  finally
    Lines.Free;
  end;
end;

procedure TResolver.LoadNetworkFromFile;
var
  IniFile: TIniFile;
  S: string;
  FullPath: string;
begin
  if not FindFileName(SNetworkFileName, FullPath) then Exit;
  IniFile := TIniFIle.Create(FullPath);
  try
    FDhcp := IniFile.ReadBool('IP', 'DHCP', False);

    S := IniFile.ReadString('IP', 'IPADDRESS', '0.0.0.0');
    FIpAddress := IP_StringToAddr(S);

    S := IniFile.ReadString('IP', 'NETMASK', '255.255.255.0');
    FNetMask := IP_StringToAddr(S);

    S := IniFile.ReadString('IP', 'GATEWAY', '0.0.0.0');
    FGateway := IP_StringToAddr(S);

    FHostName := IniFile.ReadString('DNS', 'HOSTNAME', 'localhost');

    FDomainName := IniFile.ReadString('DNS', 'DOMAINNAME', 'localdomain');

    S := IniFile.ReadString('DNS', 'NAMESERVER', '0.0.0.0');
    FNameServer := IP_StringToAddr(S);
  finally
    IniFile.Free;
  end;
end;

function TResolver.AppendDomain(const HostName: string): string;
begin
  Result := HostName;
  if DomainName <> '' then
    Result := Result + '.' + DomainName;
end;

initialization
  Resolver := TResolver.Create;

finalization
  Resolver.Free;
  Resolver := nil;

end.
