
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosTCP.pas                                           }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosTCP.pas,v 1.5 2003/10/07 19:38:19 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosTCP.pas,v $
  Revision 1.5  2003/10/07 19:38:19  iwache
  MAX_WINDOW extended to 8192.
  RETRY_TIMEOUTS are now 100, 200, 1500, 5000, 10000.
  Function GetFreePort now increments next new port.
  Procedure ProcessIP modified to fix some bugs with
  connection and disconnection behaviour.
  Function TCP_Close modified to make a linger behaviour
  (delay socket freeing until ACK from remote side).
  Function TCP_ShutDown modified:
   - wait for all data confirmed
   - Send close command and wait for confirmation
  Function TCP_Recv slightly modified. Receive data only
  when connected.
  Procedure TCP_Idle modified, also check for pending
  acknowledges.

  Revision 1.4  2003/06/12 20:21:10  iwache
  Bug in TCP_Select fixed.

  Revision 1.3  2003/06/11 21:47:45  iwache
  Some debug lines removed in finalization section.

  Revision 1.2  2003/06/11 20:54:49  iwache
  New function TCP_Select added for compatibility
  with unit Sockets.pas. Function TCP_Recv modified
  to improve thread timing. Old behaviour was: the socket
  was closed before data are readable. This is fixed.

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosTCP;

interface

uses
  Classes, WDosNetwork, WDosIP, WDosSock, WDosClock, WDosUtil,
  SysUtils, Threads;

{ Initialisation and deinitialisation }
function TCP_Init: Boolean;
function TCP_Deinit: Boolean;

{ Creation, connection & disconnection }
function TCP_Socket(Domain, SocketType, Protocol: Longint): Pointer;
function TCP_Close(Sock: Pointer): LongInt;
function TCP_Shutdown(Sock: Pointer; How: LongInt): LongInt;
function TCP_Bind(Sock: Pointer; var Addr; AddrLen: Longint): Integer;
function TCP_Listen(Sock: Pointer; MaxConnect: Longint): Integer;
function TCP_Accept(Sock: Pointer; var Addr; var Addrlen: Longint): Pointer;
function TCP_Connect(Sock: Pointer; var Addr; Addrlen: Longint): Integer;

{ Reception }
function TCP_Recv(Sock: Pointer; var Addr; AddrLen, Flags: Longint): Longint;
function TCP_RecvFrom(Sock: Pointer; var Data; DataLen, Flags: Integer; var
  AddrFrom; var FromLen: Integer): Integer;

{ Sending }
function TCP_Send(Sock: Pointer; var Addr; AddrLen, Flags: Longint): Longint;
function TCP_SendTo(Sock: Pointer; var Data; DataLen, Flags: Integer; var Addr;
  AddrLen: Integer): Integer;

{ Various }
function TCP_GetLastError: LongInt;
function TCP_ioctlsocket(Sock: Pointer; Option: LongInt; d: Pointer): LongInt;
function TCP_GetSocketName(Sock: Pointer; var Addr; var Addrlen: Longint):
  Longint;
function TCP_GetPeerName(Sock: Pointer; var Addr; var Addrlen: Longint):
  Longint;
function TCP_SetSocketOptions(Sock: Pointer; Level, OptName: Longint; var
  OptVal; OptLen: Longint): Longint;
function TCP_GetSocketOptions(Sock: Pointer; Level, OptName: Longint; var
  OptVal; var OptLen: Longint): Longint;
function TCP_Select(Sock: Pointer; var ReadReady, WriteReady, ExceptFlag: 
    Boolean): Integer;

implementation

const
  { Code Bits Flags }
  CB_FIN = 1;
  CB_SYN = 2;
  CB_RST = 4;
  CB_PSH = 8;
  CB_ACK = 16;
  CB_OOB = 32;
  { Other parameters }
  MAX_PENDING = 5;
  MTU = 1200;
  MAX_WINDOW = 8192;
  { Retry Timeouts }
  MAX_RETRY = 5;
  RETRY_TIMEOUTS: array[1..MAX_RETRY] of Integer =
  (100, 200, 1500, 5000, 10000);

type
  PTCPHeader = ^TTCPHeader;
  TTCPHeader = packed record
    SourcePort: WORD;
    TargetPort: WORD;
    SequenceNumber: DWORD;
    AckNumber: DWORD;
    HLen: BYTE;
    CodeBits: BYTE;
    Window: WORD;
    Checksum: WORD;
    UrgentPointer: WORD;
  end;

  TPseudoHeader = packed record
    SourceIP: TIPAddress;
    TargetIP: TIPAddress;
    Cero: BYTE;
    Protocol: BYTE;
    TCPLen: WORD;
  end;

  TPendingEntry = record
    LocalIP: TIPAddress;
    RemoteIP: TIPAddress;
    RemotePort: WORD;
    SequenceNumber: DWORD;
  end;

  TTCPStatus = (tcpClosed, tcpListen, tcpSYNRcvd, tcpConnected,
    tcpLocalEndWait1, tcpLocalEndWait2, tcpRemoteEndWait1,
    tcpConnecting1);

  TTCPSocket = class
    { Socket Status }
    Status: TTCPStatus;
    CanFree: Boolean;
    FreeTimeout: Int64;
    { Local & remote connection information }
    LocalIP: TIPAddress;
    LocalPort: WORD;
    RemoteIP: TIPAddress;
    RemotePort: WORD;
    { Flow Control }
    Paused: Boolean;
    NextByteToSend: DWORD;
    LastByteConfirmed: DWORD;
    LastByteAcknowledged: DWORD;
    LastByteReceived: DWORD;
    RemoteWindow: WORD;
    ACKTimeout: Int64;
    { Buffers }
    DataBuffer: AnsiString;
    { Pending connections (for listening) }
    PendingCount: BYTE;
    Pendings: array[1..MAX_PENDING] of TPendingEntry;
    { Options }
    ConnTimeout: DWORD; // in milliseconds
  end;

  TTCPFragment = class
    Socket: TTCPSocket;
    LocalIP: TIPAddress;
    TargetIP: TIPAddress;
    EndSeq: DWORD;
    Data: AnsiString;
    SentCount: BYTE;
    Timeout: Int64;
  end;

var
  SockList: TList;
  FragList: TList;
  LastError: LongInt;
  LastPort: Integer;

procedure TCP_Idle; forward;

function TCP_Checksum(SourceIP, TargetIP: TIPAddress;
  PData: PChar; Len: Word): WORD;
var
  PH: TPseudoHeader;
begin
  { Set-up the psueudo-header }
  FillChar(PH, SizeOf(PH), 0);
  PH.SourceIP := SourceIP;
  PH.TargetIP := TargetIP;
  PH.TCPLen := Swap(Word(Len));
  PH.Cero := 0;
  PH.Protocol := IPPROTO_TCP;
  { Calculate the checksum }
  TCP_Checksum := IP_Checksum(PData, Len, @PH, SizeOf(PH));
end;

function GetConnectedSocket(LocalIP, RemoteIP: TIPAddress;
  LocalPort, RemotePort: WORD): TTCPSocket;
var
  i: Integer;
begin
  for i := 0 to SockList.Count - 1 do
  begin
    Result := TTCPSocket(SockList[i]);
    if not (Result.Status in [tcpClosed, tcpListen])
      and (Result.RemoteIP = RemoteIP)
      and (Result.RemotePort = RemotePort)
      and ((Result.LocalIP = LocalIP) or (Result.LocalIP = INADDR_ANY))
      and (Result.LocalPort = LocalPort) then
      Exit;
  end;
  Result := nil;
end;

function GetAnySocket(LocalIP: TIPAddress; LocalPort: WORD): TTCPSocket;
var
  i: Integer;
begin
  for i := 0 to SockList.Count - 1 do
  begin
    Result := TTCPSocket(SockList[i]);
    if ((Result.LocalIP = LocalIP) or (Result.LocalIP = INADDR_ANY))
      and (Result.LocalPort = LocalPort) then
      Exit;
  end;
  Result := nil;
end;

function GetFreePort: Integer;
var
  I: Integer;
  Port: Integer;
begin
  // get last used port
  Port := LastPort;
  if Port = 0 then Port := 1023;

  // search for the next free port
  repeat
    Port := Port + 1;
    if Port > 8192 then Port := 1024;
    Result := Swap(Port);

    // look if the new port is used
    for I := 0 to SockList.Count - 1 do
    begin
      if TTCPSocket(SockList[i]).LocalPort = Result then
      begin
        Result := 0;
        Break;
      end;
    end;
  until (Result <> 0) or (Port = LastPort);
  if Result <> 0 then LastPort := Port;
end;

function GetPendingFragmentInfo(Socket: TTCPSocket;
  var FirstFragment: TTCPFragment; var Timeout: Int64): Integer;
var
  i: Integer;
  Entry: TTCPFragment;
begin
  Result := 0;
  Timeout := 0;
  FirstFragment := nil;
  for i := 0 to FragList.Count - 1 do
  begin
    Entry := FragList[i];
    if Entry.Socket = Socket then
    begin
      if Entry.Timeout > Timeout then
        Timeout := Entry.Timeout;
      if Result = 0 then
        FirstFragment := Entry;
      Inc(Result);
    end;
  end;
end;

procedure AddFragment(Socket: TTCPSocket; LocalIP, TargetIP: TIPAddress;
  EndSeq: DWORD; const Data: AnsiString);
var
  Fragment, Tmp: TTCPFragment;
begin
  Fragment := TTCPFragment.Create;
  Fragment.Socket := Socket;
  Fragment.LocalIP := LocalIP;
  Fragment.TargetIP := TargetIP;
  Fragment.EndSeq := EndSeq;
  Fragment.Data := Data;
  Fragment.SentCount := 1;
  GetPendingFragmentInfo(Socket, Tmp, Fragment.Timeout);
  if Fragment.Timeout = 0 then
    Fragment.Timeout := GetMSCounter + RETRY_TIMEOUTS[1];
  FragList.Add(Fragment);
end;

procedure ReSend(Fragment: TTCPFragment);
begin
  IP_Send(Fragment.LocalIP, Fragment.TargetIP, IPPROTO_TCP,
    Fragment.Data[1], Length(Fragment.Data));
end;

function Send(
  Sock: TTCPSocket;
  SourceIP, TargetIP: TIPAddress;
  SourcePort, TargetPort: WORD;
  CodeBits: BYTE;
  PData: Pointer; DataLen: WORD; ACK: DWORD): Boolean;
var
  Hdr: PTCPHeader;
  WinSize: Integer;
  StrData: AnsiString;
begin
  SetLength(StrData, SizeOf(TTCPHeader) + DataLen);
  Hdr := @StrData[1];
  { Set up the header }
  FillChar(Hdr^, SizeOf(Hdr^), 0);
  if Assigned(Sock) then
    WinSize := MAX_WINDOW - Length(Sock.DataBuffer)
  else
    WinSize := 0;
  if WinSize < 0 then
    WinSize := 0;
  Hdr^.SourcePort := SourcePort;
  Hdr^.TargetPort := TargetPort;
  if Assigned(Sock) then
  begin
    Sock.LastByteAcknowledged := Sock.LastByteReceived + 1;
    Hdr^.ACKNumber := SwapDWORD(Sock.LastByteReceived + 1);
    Hdr^.SequenceNumber := SwapDWORD(Sock.NextByteToSend);
  end
  else
  begin
    Hdr^.ACKNumber := ACK;
    Hdr^.SequenceNumber := 0;
  end;
  Hdr^.HLen := (SizeOf(TTCPHeader) div 4) shl 4;
  Hdr^.CodeBits := CodeBits;
  Hdr^.Window := Swap(Word(WinSize));
  { Add data }
  if PData <> nil then
    Move(PData^, PChar(Hdr)[SizeOf(TTCPHeader)], DataLen);
  { Calculate checksum }
  Hdr^.Checksum := 0;
  Hdr^.Checksum := TCP_Checksum(SourceIP, TargetIP, Pointer(Hdr),
    SizeOf(TTCPHeader) + DataLen);
  { Register the datagram in the fragment list }
  if DataLen > 0 then
  begin
    AddFragment(Sock, SourceIP, TargetIP,
      SwapDWORD(Hdr^.SequenceNumber) + DataLen, StrData);
  end;
  { Send the Datagram }
  Send := IP_Send(SourceIP, TargetIP, IPPROTO_TCP,
    Hdr^, SizeOf(TTCPHeader) + DataLen) = 0;
  //  WriteLn('***** Se envi el paquete    ', SwapLong(Hdr^.SequenceNumber));
  //  WriteLn('***** Se confirm el paquete ', Sockets[Socket].LastByteReceived);
      { if data was sent, increase the Secuence Number }
  if Assigned(Sock) then
    Inc(Sock.NextByteToSend, DataLen);
end;

procedure ProcessIP(SourceIP, TargetIP: TIPAddress;
  var Data; DataLen: WORD);
var
  CodeBits: BYTE;
  Hdr: PTCPHeader;
  Sock: TTCPSocket;
  SourcePort, TargetPort: WORD;
  HLen, nPend, AntLen: Integer;
  SequenceNumber, AcknowledgeNumber, ExpectedByte: DWORD;
begin
  Hdr := @Data;
  if Hdr.TargetPort = 0 then
    Exit;
  HLen := BYTE(Hdr.HLen shr 4) * 4;
  CodeBits := Hdr.CodeBits and 63;
  SourcePort := Hdr.SourcePort;
  TargetPort := Hdr.TargetPort;
  SequenceNumber := SwapDWORD(Hdr.SequenceNumber);
  AcknowledgeNumber := SwapDWORD(Hdr.AckNumber);

  { Debug }
//  WriteLn('  *TCP*');
//  Writeln('   Source IP      : ', IP_AddrToString(SourceIP));
//  Writeln('   TargetIP       : ', IP_AddrToString(TargetIP));
//  WriteLn('   Source Port    : ', Swap(SourcePort));
//  WriteLn('   Target Port    : ', Swap(TargetPort));
//  WriteLn('   HLen           : ', HLen);
//  WriteLn('   DataLen        : ', DataLen - HLen);
//  WriteLn('   Code Bits      : ', CodeBits);
//  WriteLn('   Sequence Number: ', SequenceNumber);
//  WriteLn('   ACK Number     : ', AcknowledgeNumber);
//  WriteLn('   Window         : ', Swap(Hdr.Window));
//  WriteLn('   Checksum       : ', IntToHex(Hdr.Checksum, 4));
//  Hdr.Checksum := 0;
//  WriteLn('   Checksum Recalc: ',
//  IntToHex(TCP_Checksum(SourceIP, TargetIP, @Data, DataLen), 4));
//  WriteLn('   Urgent Pointer : ', Swap(Hdr.UrgentPointer));

  { Se recibi un datagrama TCP. Primero que nada hay que ver si existe  }
  { algn socket conectado listo para recibir este datagrama.            }
  Sock := GetConnectedSocket(TargetIP, SourceIP, TargetPort, SourcePort);

  { Si lo anterior no result, buscamos algn socket no conectado, pero  }
  { listo para recibir este datagrama.                                   }
  if Sock = nil then
    Sock := GetAnySocket(TargetIP, TargetPort);

  { Verificamos el nmero de Secuencia del Emisor. Si no se corresponde  }
  { con lo que debera, nos vamos, ignorando el Paquete. La nica        }
  { excepcin a esto es el paquete de SYN, que es el primero.            }
  if (Sock <> nil) and not (Sock.Status in [tcpListen, tcpConnecting1]) then
  begin
    ExpectedByte := Sock.LastByteReceived + 1;
    if (SequenceNumber = ExpectedByte) or (SequenceNumber = 0) then
    begin
      // sequence number is correct
      Inc(Sock.LastByteReceived, DWORD(DataLen - HLen));
      { Si llegaron datos, los almacenamos el el Buffer del Socket }
      DataLen := DataLen - HLen;
      if (Sock <> nil) and (DataLen > 0) then
      begin
        { Agrandamos el Buffer de Datos del Socket }
        AntLen := Length(Sock.DataBuffer);
        SetLength(Sock.DataBuffer, Length(Sock.DataBuffer) + DataLen);
        { Copiamos los Nuevos Datos }
        Move(PChar(@Data)[HLen], Sock.DataBuffer[AntLen + 1], DataLen);
      end;
    end;
//    else
//    begin
//      // Nmero de secuencia incorrecto
//      if not (Sock.Status in [tcpLocalEndWait1, tcpLocalEndWait2,
//        tcpRemoteEndWait1]) then
//      begin
//        // WriteLn('Se RECHAZO el Paquete ', SequenceNumber,
//        //     ' (se esperaba el ', ExpectedByte, '). CodeBits: ', CodeBits);
//        if (CodeBits and (CB_FIN or CB_SYN)) <> 0 then
//        begin
//          //  WriteLn('CB_END / CB_SYN');
//          //  Send(Sock, TargetIP, SourceIP, TargetPort,
//          //      SourcePort, CB_RST, nil, 0);
//        end
//        else
//        begin
//          Send(Sock, TargetIP, SourceIP, TargetPort,
//            SourcePort, CB_ACK, nil, 0, 0);
//        end;
//        Exit;
//      end;
//    end;
  end;

  { Procesamiento para Code Bits = CB_SYN (Conexin entrante)            }
  if (CodeBits and CB_SYN) = CB_SYN then
  begin
    // WriteLn('SYN!!!');
    { Si hay un socket en modo listen para este paquete, aceptamos     }
    { la conexin. De lo contrario, la rechazamos.                     }
    if Sock = nil then
    begin
      Send(nil, TargetIP, SourceIP, TargetPort, SourcePort, CB_RST or CB_ACK,
        nil, 0, SwapDWORD(SwapDWORD(Hdr^.SequenceNumber) + 1));
    end
    else
      case Sock.Status of
        tcpListen:
          if (Sock.PendingCount < MAX_PENDING) then
          begin
            { Se acepta la conexin. La ponemos en la cola de      }
            { conexiones para ser aceptadas.                       }
            Inc(Sock.PendingCount);
            nPend := Sock.PendingCount;
            Sock.Pendings[nPend].LocalIP := TargetIP;
            Sock.Pendings[nPend].RemoteIP := SourceIP;
            Sock.Pendings[nPend].RemotePort := SourcePort;
            Sock.Pendings[nPend].SequenceNumber := SequenceNumber;
          end
          else
          begin
            Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort, CB_RST, nil,
              0, 0);
          end;
        tcpConnecting1:
          begin
            Sock.LastByteReceived := SequenceNumber;
            Sock.NextByteToSend := AcknowledgeNumber;
            Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
              CB_ACK, nil, 0, 0);
            Sock.Status := tcpConnected;
          end;
      else
        begin
          Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort, CB_RST, nil, 0,
            0);
        end;
      end;
  end;

  { Procesamiento para CB_END                                            }
  if (CodeBits and CB_FIN) <> 0 then
  begin
    // WriteLn('END!!!!');
    if Sock <> nil then
      case Sock.Status of
        tcpLocalEndWait1:

          // #iwa-2003-10-01: What happens when both sides close simultanious?

          if CodeBits = CB_FIN then
          begin
            // both ends try to close connection at the same time
            Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort, CB_ACK, nil,
                0, 0);
            Sock.Status := tcpClosed;
            Sock.ACKTimeout := 0;
          end;
        tcpLocalEndWait2:
          begin
            // Acknowledge the FIN signal
            Sock.NextByteToSend := AcknowledgeNumber;
            Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
              CB_ACK, nil, 0, 0);
            Sock.Status := tcpClosed;
            Sock.ACKTimeout := 0;
          end;
        tcpConnected:
          begin
            // accept FIN only when all pending data are received
            ExpectedByte := Sock.LastByteReceived + 1;
            if SequenceNumber = ExpectedByte then
            begin
              // WriteLn('Peticin de cierre remoto');
              Inc(Sock.LastByteReceived);
              Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
                CB_ACK, nil, 0, 0);
              Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
                CB_FIN or CB_ACK, nil, 0, 0);
              Sock.Status := tcpRemoteEndWait1;
              Sock.ACKTimeout := 0;
            end;
          end;
      end;
  end;

  { Procesamiento para CB_RST }
  if (CodeBits and CB_RST) = CB_RST then
  begin
    // WriteLn('RST!!!! ', Swap(SourcePort), ' ', Swap(TargetPort));
    { Si no es para nadie, nos vamos }
    if (Sock <> nil) and not (Sock.Status in [tcpListen]) then
    begin
      Sock.Status := tcpClosed;
      Sock.ACKTimeout := 0;
    end;
  end;

  { Procesamiento de ACKs. Bsicamente no hacemos ms que almacenar el   }
  { Nmero de Confirmacin en el Socket elegido.                         }
  if (CodeBits and CB_ACK) = CB_ACK then
  begin
    // WriteLn('ACK!!!!');
    if Sock <> nil then
    begin
      case Sock.Status of
        tcpSynRcvd:
          begin
            Sock.Status := tcpConnected;
          end;
        tcpLocalEndWait1:
          begin
            // accept FIN only when all pending data are received
            ExpectedByte :=  Sock.NextByteToSend + 1;
            if AcknowledgeNumber = ExpectedByte then
            begin
              Inc(Sock.LastByteReceived);
              Sock.Status := tcpLocalEndWait2;
            end;
          end;
        tcpRemoteEndWait1:
          if CodeBits = CB_ACK then
          begin
            // WriteLn('Socket cerrado remotamente !');
            Sock.Status := tcpClosed;
            Sock.ACKTimeout := 0;
          end;
        tcpConnecting1:
          begin
            Sock.LastByteReceived := SequenceNumber;
            Sock.NextByteToSend := AcknowledgeNumber;
            Send(Sock, TargetIP, SourceIP, TargetPort,
              SourcePort, CB_RST or CB_ACK, nil, 0, 0);
          end;
      end;
      if AcknowledgeNumber > Sock.LastByteConfirmed then
        Sock.LastByteConfirmed := AcknowledgeNumber;
    end;
  end;

  { Anotamos el tamao de la ventana remota }
  if (Sock <> nil) then
  begin
    Sock.RemoteWindow := Swap(Hdr.Window);
    // WriteLn('RemoteWindow: ', Sock.RemoteWindow);
  end;
end;

function TCP_Init: Boolean;
var
  PS: TProtocolStruct;
begin
  { Add the TCP processor to de IP Stack }
  Result := IP_AddProcessor(IPPROTO_TCP, @ProcessIP);
  {}
  PS.Domain := PF_INET;
  PS.SocketType := SOCK_STREAM;
  PS.Protocol := IPPROTO_IP;
  PS.FNSocket := TCP_Socket;
  PS.FNClose := TCP_Close;
  PS.FNGetLastError := TCP_GetLastError;
  PS.FNSend := TCP_Send;
  PS.FNSendTo := TCP_SendTo;
  PS.FNRecv := TCP_Recv;
  PS.FNRecvFrom := TCP_RecvFrom;
  PS.FNBind := TCP_Bind;
  PS.FNListen := TCP_Listen;
  PS.FNAccept := TCP_Accept;
  PS.FNConnect := TCP_Connect;
  PS.FNShutdown := TCP_Shutdown;
  PS.FNGetSocketName := TCP_GetSocketName;
  PS.FNGetPeerName := TCP_GetPeerName;
  PS.FNSetSockOpt := TCP_SetSocketOptions;
  PS.FNGetSockOpt := TCP_GetSocketOptions;
  PS.FNioctlsocket := TCP_ioctlsocket;
  PS.FNSelect := TCP_Select;
  Result := Result and AddProtocol(PS);
  PS.Protocol := IPPROTO_TCP;
  Result := Result and AddProtocol(PS);
end;

function TCP_GetLastError: LongInt;
begin
  TCP_GetLastError := LastError;
end;

function TCP_Deinit: Boolean;
begin
  TCP_Deinit := IP_RemoveProcessor(IPPROTO_TCP);
end;

function TCP_Socket(Domain, SocketType, Protocol: Longint): Pointer;
var
  SockObj: TTCPSocket;
begin
  { Initialise new socket }
  SockObj := TTCPSocket.Create;
  SockObj.Status := tcpClosed;
  SockObj.CanFree := False;
  SockObj.LocalIP := 0;
  SockObj.LocalPort := 0;
  SockObj.RemoteIP := 0;
  SockObj.RemotePort := 0;
  SockObj.Paused := False;
  SockObj.NextByteToSend := Random($0FFFFF);
  SockObj.LastByteConfirmed := 0;
  SockObj.LastByteAcknowledged := 0;
  SockObj.LastByteReceived := 0;
  SockObj.DataBuffer := '';
  SockObj.PendingCount := 0;
  SockObj.ACKTimeout := 0;
  { Default Options }
  SockObj.ConnTimeout := 5000;
  { Add new socket to the socket list }
  SockList.Add(SockObj);
  { Done }
  TCP_Socket := SockObj;
end;

procedure TCP_Free(Sock: TTCPSocket);
var
  i: Integer;
  Fragment: TTCPFragment;
begin
  { If connected, reset the connection }
  if not (Sock.Status in [tcpListen, tcpClosed]) then
  begin
    Send(Sock, Sock.LocalIP, Sock.RemoteIP, Sock.LocalPort,
      Sock.RemotePort, CB_RST, nil, 0, 0);
  end;
  { Remove any pending fragments }
  i := 0;
  while i < FragList.Count do
  begin
    Fragment := FragList[i];
    if Fragment.Socket = Sock then
    begin
      FragList.Remove(Fragment);
      Fragment.Free;
    end
    else
    begin
      Inc(i);
    end;
  end;
  { Free the socket Structure, and remove it from the socket list }
  SockList.Remove(Sock);
  Sock.Free;
end;

function TCP_Close(Sock: Pointer): LongInt;
var
  SockObj: TTCPSocket;
begin
  SockObj := TTCPSocket(Sock);
  { If the socket is connected, we disconnect it before proceeding }
  if SockObj.Status = tcpConnected then
  begin
    // Send close command
    Send(SockObj, SockObj.LocalIP, SockObj.RemoteIP, SockObj.LocalPort,
      SockObj.RemotePort, CB_FIN or CB_ACK, nil, 0, 0);
    Inc(SockObj.NextByteToSend);
    SockObj.Status := tcpLocalEndWait1;
    // Mark socket as "Freeable"
    SockObj.CanFree := True;
    SockObj.FreeTimeout := GetMSCounter + 5000;
  end

  // delay socket freeing until ACK from remote side 
  else if SockObj.Status in [tcpLocalEndWait2, tcpRemoteEndWait1] then
  begin
    // Mark socket as "Freeable"
    SockObj.CanFree := True;
    SockObj.FreeTimeout := GetMSCounter + 5000;
  end

  else
  begin
    TCP_Free(SockObj);
  end;
  { Listo }
  TCP_Close := 0;
end;

function TCP_ShutDown(Sock: Pointer; How: LongInt): LongInt;
var
  Timeout: Int64;
  SockObj: TTCPSocket;
  Retry: Integer;
  Done: Boolean;
begin
  SockObj := TTCPSocket(Sock);
  { Check Socket status }
  if SockObj.Status <> tcpConnected then
  begin
    TCP_ShutDown := -1;
    LastError := WSAENOTCONN;
    Exit;
  end

  else if SockObj.Status = tcpRemoteEndWait1 then
  begin
    TCP_ShutDown := 0;
    Exit;
  end;

  // wait for all data confirmed
  Timeout := GetMSCounter + RETRY_TIMEOUTS[MAX_RETRY];
  repeat
    NET_Work_Low;
    Yield;
    if SockObj.LastByteConfirmed = SockObj.NextByteToSend then Break;
  until GetMSCounter > Timeout;

  // Send close command and wait for confirmation.
  Retry := 1;
  SockObj.Status := tcpLocalEndWait1; // FIN_WAIT_1
  //  waiting for FIN_WAIT_2 or CLOSED
  repeat
    Send(SockObj, SockObj.LocalIP, SockObj.RemoteIP, SockObj.LocalPort,
        SockObj.RemotePort, CB_FIN or CB_ACK, nil, 0, 0);
    Timeout := GetMSCounter + RETRY_TIMEOUTS[Retry];
    repeat
      NET_Work_Low;
      Yield;
      Done := (SockObj.Status in [tcpLocalEndWait2, tcpClosed]);
    until (GetMSCounter > Timeout) or Done;
    if Done then Break;
    Dec(Retry);
  until Retry = MAX_RETRY;

  { Done }
  if Done then
  begin
    TCP_ShutDown := 0;
  end
  else
  begin
    TCP_ShutDown := -1;
    LastError := WSAETIMEDOUT;
  end;
end;

function TCP_Bind(Sock: Pointer; var Addr; AddrLen: Longint): Integer;
var
  SockObj: TTCPSocket;
  InAddr: TSockAddr;
begin
  TCP_Bind := SOCKET_ERROR;
  SockObj := TTCPSocket(Sock);
  InAddr := TSockAddr(Addr);
  // Check the given Address
  if InAddr.sin_family <> AF_INET then
  begin
    LastError := WSAEFAULT;
    Exit;
  end;
  // If it's necessary, we choose a free local port for the socket
  if InAddr.sin_port = 0 then
    InAddr.sin_port := GetFreePort;
  // assign the IP and port number to socket
  SockObj.LocalIP := InAddr.sin_addr.S_addr;
  SockObj.LocalPort := InAddr.sin_port;
  { Listo }
  TCP_Bind := 0;
end;

function TCP_Listen(Sock: Pointer; MaxConnect: Longint): Integer;
var
  SockObj: TTCPSocket;
begin
  SockObj := TTCPSocket(Sock);
  { If the socket is connected or in the process of getting connected, }
  { we cannot put it in passive mode, so we quit.                      }
  if SockObj.Status <> tcpClosed then
  begin
    TCP_Listen := SOCKET_ERROR;
    LastError := WSAEISCONN;
    Exit;
  end;
  { If the socket is not bound to a valid Address and Port, we quit    }
  if (SockObj.LocalIP = INADDR_NONE) or (SockObj.LocalPort = 0) then
  begin
    TCP_Listen := SOCKET_ERROR;
    LastError := WSAEINVAL;
    Exit;
  end;
  { Done }
  SockObj.Status := tcpListen;
  TCP_Listen := 0;
end;

function TCP_Select(Sock: Pointer; var ReadReady, WriteReady, ExceptFlag:
    Boolean): Integer;
var
  SockObj: TTCPSocket;
begin
  ReadReady := False;
  WriteReady := False;
  ExceptFlag := False;

  SockObj := TTCPSocket(Sock);
  // If the socket is listening  ...
  if SockObj.Status = tcpListen then
  begin
    { If there's no incomming connection to accept, we quit }
    if SockObj.PendingCount <= 0 then
    begin
      TCP_Select := 0;
    end
    else
    begin
      ReadReady := True;
      TCP_Select := 0;
    end;
  end
  else
  // if the socket is not listening ...
  begin
    // If there's nothing to read, and the socket is not connected, we
    // should return false. However, if there's something to
    // read, we should do it even if it's not connected.
    if (Length(SockObj.DataBuffer) > 0) then
    begin
      ReadReady := True;
      TCP_Select := 0;
    end
    else
    begin
      TCP_Select := 0;
      if not (SockObj.Status in [tcpConnected,
          tcpLocalEndWait1, tcpLocalEndWait2]) then
      begin
        ReadReady := True;
      end;
    end;
  end;
end;

function TCP_Accept(Sock: Pointer; var Addr; var Addrlen: Longint): Pointer;
var
  i: Integer;
  Timeout: Int64;
  SockObj, NewSock: TTCPSocket;
begin
  SockObj := TTCPSocket(Sock);
  { If the socket is not listening, we quit }
  if SockObj.Status <> tcpListen then
  begin
    TCP_Accept := nil;
    LastError := WSAEINVAL;
    Exit;
  end;
  { If there's not enough to copy the remote peer's address and port, }
  { we quit.                                                          }
  if AddrLen < SizeOf(TSockAddr) then
  begin
    TCP_Accept := nil;
    LastError := WSAEMSGSIZE;
    Exit;
  end;
  { If there's no incomming connection to accept, we quit }
  if SockObj.PendingCount <= 0 then
  begin
    TCP_Accept := nil;
    LastError := WSAEWOULDBLOCK;
    Exit;
  end;
  { All Right. We create a new socket for the incomming connection }
  NewSock := TCP_Socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
  if NewSock = nil then
  begin
    TCP_Accept := nil;
    Exit;
  end;
  { The new socket has to be bound to the same local address and port }
  { as the listening socket. The remote IP and port of the new socket }
  { must be the ones from the connecting socket.                      }
  NewSock.LocalIP := SockObj.Pendings[1].LocalIP;
  NewSock.RemoteIP := SockObj.Pendings[1].RemoteIP;
  NewSock.LocalPort := SockObj.LocalPort;
  NewSock.RemotePort := SockObj.Pendings[1].RemotePort;
  NewSock.LastByteReceived := SockObj.Pendings[1].SequenceNumber;
  NewSock.LastByteAcknowledged := 0;
  { Remove the Accepted connection from the pending queue }
  Dec(SockObj.PendingCount);
  for i := 1 to SockObj.PendingCount do
  begin
    SockObj.Pendings[i] := SockObj.Pendings[i + 1];
  end;
  { Change the status of the new socket  }
  NewSock.Status := tcpSynRcvd;
  { Send the first confirmation to the remote end }
  Send(NewSock, NewSock.LocalIP, NewSock.RemoteIP, NewSock.LocalPort,
    NewSock.RemotePort, CB_SYN or CB_ACK, nil, 0, 0);
  { Wait for the connection to complete }
  Timeout := GetMSCounter + 5000; { 5 seconds }
  repeat
    NET_Work_Low;
    Yield;
  until (GetMSCounter > Timeout) or (NewSock.Status in [tcpConnected,
    tcpClosed]);
  { If the connection was completed succesfully, we return the new  }
  { socket. Otherwise, we return nil, and the newly created socket  }
  { is destroyed.                                                   }
  if NewSock.Status = tcpConnected then
  begin
    { Copy the remote Peer's address to the user-supplied buffer }
    NewSock.NextByteToSend := NewSock.LastByteConfirmed;
    TCP_GetPeerName(NewSock, Addr, AddrLen);
    TCP_Accept := NewSock;
  end
  else
  begin
    TCP_Accept := nil;
    LastError := WSAETIMEDOUT;
    TCP_Close(NewSock);
  end;
end;

function TCP_Send(Sock: Pointer; var Addr; AddrLen, Flags: Longint): Longint;

  procedure TestRemoteWindow(SockObj: TTCPSocket; Data: Pointer; Len: Integer);
  var
    Tmp: Int64;
    FE: TTCPFragment;
  begin
    if GetPendingFragmentInfo(SockObj, FE, Tmp) > 0 then
    begin
      // Send first pending fragment
      ReSend(FE);
    end
    else
    begin
      // Send a new fragment
      Send(SockObj, SockObj.LocalIP, SockObj.RemoteIP,
          SockObj.LocalPort, SockObj.RemotePort,
          CB_ACK or CB_PSH, Data, Len, 0);
    end;
  end;

var
  P: PChar;
  FragLen: DWORD;
  SockObj: TTCPSocket;
  TmpFrag: TTCPFragment;
  Timeout, T2, TmpInt64: Int64;
begin
  SockObj := TTCPSocket(Sock);

  { if the Socket is not connected, we quit }
  if SockObj.Status <> tcpConnected then
  begin
    TCP_Send := SOCKET_ERROR;
    LastError := WSAENOTCONN;
    Exit;
  end;

  { All Right. Send the Data }
  P := @Addr;
  TCP_Send := AddrLen;
  while (AddrLen > 0) and (SockObj.Status = tcpConnected) do
  begin

    { Calculate this fragment's size }
    if AddrLen > MTU then
      FragLen := MTU
    else
      FragLen := AddrLen;

    { Wait for the other end to be ready to receive one fragment }
    SockObj.Paused := True;
    T2 := GetMSCounter + 1000;
    Timeout := GetMSCounter + 60000;
    while (SockObj.RemoteWindow < FragLen)
      and (SockObj.Status = tcpConnected)
      and (GetMSCounter < Timeout) do
    begin
      NET_Work_Low;
      TCP_Idle;
      Yield;
      if GetMSCounter > T2 then
      begin
        TestRemoteWindow(SockObj, P, FragLen);
        T2 := GetMSCounter + 1000;
      end;
    end;
    SockObj.Paused := False;
    if (SockObj.RemoteWindow < FragLen) or (SockObj.Status <> tcpConnected) then
      Break;

    { Wait for the fragment queue to have not more than 5 fragments }
    while (GetPendingFragmentInfo(SockObj, TmpFrag, TmpInt64) > 5)
      and (SockObj.Status = tcpConnected) do
    begin
      //			WriteLn('Esperando a que se enven los fragmentos ',
      //            GetPendingFragmentInfo(PSock, nil, nil));
      Net_Work_Low;
      TCP_Idle;
      Yield;
    end;
    if SockObj.Status <> tcpConnected then
      break;

    { Send the Fragment }
    if not Send(SockObj, SockObj.LocalIP, SockObj.RemoteIP, SockObj.LocalPort,
      SockObj.RemotePort, CB_ACK or CB_PSH, P, FragLen, 0) then
      break;

    { Promote TCP Processing }
    Net_Work_Low;
    TCP_Idle;
    Yield;

    { Go for next fragment }
    Dec(AddrLen, LongInt(FragLen));
    Inc(P, FragLen);
  end;

  // Check to see if everything could be sent. Otherwise, close the
  // socket and return an error condition.
  if AddrLen <> 0 then
  begin
    TCP_Send := SOCKET_ERROR;
    LastError := WSAECONNABORTED;
    if SockObj.Status = tcpConnected then
    begin
      // WriteLn('Socket cerrado por error en Send!');
      SockObj.Status := tcpClosed;
      SockObj.ACKTimeout := 0;
    end;
  end;
end;

function TCP_SendTo(Sock: Pointer; var Data; DataLen, Flags: Integer; var Addr;
  AddrLen: Integer): Integer;
begin
  Result := SOCKET_ERROR;
  LastError := WSAEOPNOTSUPP;
end;

function TCP_Connect(Sock: Pointer; var Addr; Addrlen: Longint): Integer;
var
  i: Integer;
  Timeout: Int64;
  Retries: Integer;
  SockObj: TTCPSocket;
  Gateway: TIPAddress;
  InAddr: TSockAddr;
begin
  SockObj := TTCPSocket(Sock);
  InAddr := TSockAddr(Addr);
  { Check the given Address }
  if InAddr.sin_family <> AF_INET then
  begin
    TCP_Connect := SOCKET_ERROR;
    LastError := WSAEFAULT;
    Exit;
  end;
  { If the socket is already connected, we quit }
  if SockObj.Status <> tcpClosed then
  begin
    TCP_Connect := SOCKET_ERROR;
    LastError := WSAEISCONN;
    Exit;
  end;
  { If the socket is not bound to a local IP, now it's the time to do it }
  if (SockObj.LocalIP = INADDR_ANY) or (SockObj.LocalIP = 0)
    or (SockObj.LocalIP = INADDR_BROADCAST) then
  begin
    if not IP_GetRouteForTarget(InAddr.sin_addr.S_addr, SockObj.LocalIP, Gateway)
      then
    begin
      TCP_Connect := SOCKET_ERROR;
      LastError := WSAENETUNREACH;
      Exit;
    end;
  end;
  { If the socket is not bound to a local port, now it's the time to do it }
  if SockObj.LocalPort = 0 then
    SockObj.LocalPort := GetFreePort;
  //	WriteLn('Local Port: ', Swap(PSock.LocalPort));
      { Bind the socket to the remote peer }
  SockObj.RemoteIP := InAddr.sin_addr.S_addr;
  SockObj.RemotePort := InAddr.sin_port;
  SockObj.NextByteToSend := 1000;
  SockObj.LastByteReceived := $FFFFFFFF; // -1
  SockObj.LastByteAcknowledged := 0;
  SockObj.Status := tcpConnecting1;
  { The retries are given by the ConnTimeout parameter }
  Retries := SockObj.ConnTimeout div 500;
  if Retries < 1 then
    Retries := 1;
  for i := 1 to Retries do
  begin
    { Send a request for connection. }
    Send(SockObj, SockObj.LocalIP, SockObj.RemoteIP, SockObj.LocalPort,
      SockObj.RemotePort, CB_SYN, nil, 0, 0);
    { Wait for Ack }
    Timeout := GetMSCounter + 500; { 1/2 second }
    repeat
      NET_Work_Low;
      Yield;
    until (GetMSCounter > Timeout) or (SockObj.Status in [tcpConnected,
      tcpClosed]);
    if SockObj.Status <> tcpConnecting1 then
      Break;
  end;
  { Did the connection end succesfully ? }
  case SockObj.Status of
    tcpConnected:
      begin
        SockObj.NextByteToSend := SockObj.LastByteConfirmed;
        TCP_Connect := 0;
      end;
    tcpClosed:
      begin
        TCP_Connect := SOCKET_ERROR;
        LastError := WSAECONNREFUSED;
      end;
  else
    begin
      SockObj.Status := tcpClosed;
      TCP_Connect := SOCKET_ERROR;
      LastError := WSAETIMEDOUT;
    end;
  end;
end;

function TCP_Recv(Sock: Pointer; var Addr; AddrLen, Flags: Longint): Longint;
var
  SockObj: TTCPSocket;
begin
  NET_Work_Low;
  TCP_Idle;
  SockObj := TTCPSocket(Sock);
  { If there's nothing to read, and the socket is not connected, we }
     { should return an error code. However, if there's something to   }
     { read, we should do it even if it's not connected.               }
  if (Length(SockObj.DataBuffer) = 0) then
  begin
    if not (SockObj.Status in [tcpConnected, tcpLocalEndWait1, tcpLocalEndWait2])
      then
    begin
      TCP_Recv := SOCKET_ERROR;
      LastError := WSAENOTCONN;
    end
    else if AddrLen > 0 then
    begin
      TCP_Recv := SOCKET_ERROR;
      LastError := WSAEWOULDBLOCK;
    end
    else
    begin
      TCP_Recv := 0;
    end;
    Exit;
  end;
  { AddrLen = 0? }
  if AddrLen <= 0 then
  begin
    TCP_Recv := 0;
    Exit;
  end;
  { If there's not enough data to be read, we read only what we have }
  if AddrLen > Length(SockObj.DataBuffer) then
    AddrLen := Length(SockObj.DataBuffer);
  { Read data into the user-supplied buffer }
  if AddrLen > 0 then
    Move(SockObj.DataBuffer[1], Addr, AddrLen);
  { If necessary, remove the fetched data from the buffer }
  if (AddrLen > 0) and ((Flags and MSG_PEEK) <> MSG_PEEK) then
  begin
    SockObj.DataBuffer := Copy(SockObj.DataBuffer, AddrLen + 1,
      Length(SockObj.DataBuffer));
    { The reception window has changed, because more room is available }
    { now. Schedule to tell the other end in at least 10 ms.           }

    if SockObj.Status = tcpConnected then
      if AddrLen > 1 then
      begin
        if (SockObj.LastByteReceived + 1) > SockObj.LastByteAcknowledged then
          Send(SockObj, SockObj.LocalIP, SockObj.RemoteIP, SockObj.LocalPort,
              SockObj.RemotePort, CB_ACK, nil, 0, 0);
        SockObj.ACKTimeout := 0;
      end
      else
      begin
        SockObj.ACKTimeout := GetMSCounter + 10;
      end;
  end;
  TCP_Recv := AddrLen;
end;

function TCP_RecvFrom(Sock: Pointer; var Data; DataLen, Flags: Integer; var
  AddrFrom; var FromLen: Integer): Integer;
begin
  Result := SOCKET_ERROR;
  LastError := WSAEOPNOTSUPP;
end;

function TCP_ioctlsocket(Sock: Pointer; Option: LongInt; d: Pointer): LongInt;
var
  dw: DWORD;
  SockObj: TTCPSocket;
begin
  NET_Work_Low;
  TCP_Idle;
  TCP_ioctlsocket := 0;
  SockObj := TTCPSocket(Sock);
  case Option of
    FIONREAD:
      begin
        dw := Length(SockObj.DataBuffer);
        Move(dw, d^, 4);
      end;
  else
    begin
      TCP_ioctlsocket := SOCKET_ERROR;
      LastError := WSAEINVAL;
    end;
  end;
end;

function TCP_GetSocketName(Sock: Pointer; var Addr; var Addrlen: Longint):
  Longint;
var
  SockObj: TTCPSocket;
begin
  SockObj := TTCPSocket(Sock);
  { Check to see if there's enough room in the destination buffer }
  if AddrLen < SizeOf(TSockAddr) then
  begin
    TCP_GetSocketName := SOCKET_ERROR;
    LastError := WSAEMSGSIZE;
    Exit;
  end;
  { Copy the local IP and Port to the destination buffer }
  TSockAddr(Addr).sin_family := AF_INET;
  TSockAddr(Addr).sin_addr.S_addr := SockObj.LocalIP;
  TSockAddr(Addr).sin_port := SockObj.LocalPort;
  AddrLen := SizeOf(TSockAddr);
  { Done }
  TCP_GetSocketName := 0;
end;

function TCP_GetPeerName(Sock: Pointer; var Addr; var Addrlen: Longint):
  Longint;
var
  SockObj: TTCPSocket;
begin
  SockObj := TTCPSocket(Sock);
  { Check to see if there's enough room in the destination buffer }
  if AddrLen < SizeOf(TSockAddr) then
  begin
    TCP_GetPeerName := SOCKET_ERROR;
    LastError := WSAEMSGSIZE;
    Exit;
  end;
  { Check to see if the socket is connected }
  if (SockObj.Status <> tcpConnected) then
  begin
    TCP_GetPeerName := SOCKET_ERROR;
    LastError := WSAENOTCONN;
    Exit;
  end;
  { Copy the remote IP and Port to the destination buffer }
  TSockAddr(Addr).sin_family := AF_INET;
  TSockAddr(Addr).sin_addr.S_addr := SockObj.RemoteIP;
  TSockAddr(Addr).sin_port := SockObj.RemotePort;
  AddrLen := SizeOf(TSockAddr);
  { Done }
  TCP_GetPeerName := 0;
end;

function TCP_SetSocketOptions(Sock: Pointer; Level, OptName: Longint; var
  OptVal; OptLen: Longint): Longint;
var
  PLong: ^LongInt;
  SockObj: TTCPSocket;
begin
  Result := SOCKET_ERROR;
  SockObj := TTCPSocket(Sock);
  PLong := @OptVal;
  if Level = SOL_SOCKET then
  begin
    case OptName of
      SO_CONNTIMEOUT:
        if OptLen = SizeOf(LongInt) then
        begin
          SockObj.ConnTimeout := PLong^;
          Result := 0;
        end;
    end;
  end;
end;

function TCP_GetSocketOptions(Sock: Pointer; Level, OptName: Longint; var
  OptVal; var OptLen: Longint): Longint;
var
  PLong: ^LongInt;
  SockObj: TTCPSocket;
begin
  Result := SOCKET_ERROR;
  SockObj := TTCPSocket(Sock);
  PLong := @OptVal;
  if Level = SOL_SOCKET then
  begin
    case OptName of
      SO_CONNTIMEOUT:
        if OptLen = SizeOf(LongInt) then
        begin
          PLong^ := SockObj.ConnTimeout;
          Result := 0;
        end;
    end;
  end;
end;

procedure TCP_Idle;
var
  Ahora: Int64;
  S: TTCPSocket;
  i, j: Integer;
  Timeout: Int64;
  Fragment: TTCPFragment;
begin
  // Check the fragment list to determine if any fragment needs to be
  // re-sent or deleted from the list.
  i := 0;
  Ahora := GetMSCounter;
  while i < FragList.Count do
  begin
    Fragment := TTCPFragment(FragList[i]);
    S := Fragment.Socket;

    if (Fragment.EndSeq <= S.LastByteConfirmed) then
    begin
      // The fragment has been acknowledged. Remove it from the list.
      FragList.Remove(Fragment);
      Fragment.Free;

    end
    else if (not S.Paused) and (Fragment.SentCount >= MAX_RETRY)
      and (Ahora > Fragment.Timeout) then
    begin
      // Enough retries have been made. Mark the socket as closed.
      S.Status := tcpClosed;
      S.ACKTimeout := 0;
      FragList.Remove(Fragment);
      Fragment.Free;
      //			WriteLn('Socket cerrado por exceso de reenvos');

    end
    else if (S.Status = tcpClosed) then
    begin
      // The socket has been closed. Remove the fragment, since
      // it's useless now.
      FragList.Remove(Fragment);
      Fragment.Free;

    end
    else if (S.Status = tcpConnected) and (not S.Paused)
      and (Ahora > Fragment.Timeout) then
    begin
      // Time to retry
//			WriteLn(Fragment.EndSeq, ' ', S.LastByteConfirmed);
//			WriteLn('Retry ', Fragment.SentCount,
//			  ' of fragment ', i, ' of socket ', Integer(S));
      ReSend(Fragment);
      Inc(Fragment.SentCount);
      Timeout := Ahora + RETRY_TIMEOUTS[Fragment.SentCount];
      Fragment.Timeout := Timeout;
      // Also increase timeouts for all other fragments of this socket
      for j := i + 1 to FragList.Count - 1 do
      begin
        Fragment := TTCPFragment(FragList[j]);
        if Fragment.Socket = S then
          Fragment.Timeout := Timeout;
      end;
      // Next Fragment
      Inc(i);

    end
    else
    begin
      // Next Fragment
      Inc(i);

    end;

  end;

  // Check for sockets that can be freed because they completed the
  // close sequence.
  // Also check for pending acknowledges
  i := 0;
  while i < SockList.Count do
  begin
    S := TTCPSocket(SockList[i]);
    if S.CanFree and ((S.Status = tcpClosed) or (S.FreeTimeout < Ahora)) then
    begin
      TCP_Free(S);
    end
    else
    begin
      if (S.Status = tcpConnected) and (S.ACKTimeout <> 0)
          and (S.AckTimeout < Ahora) then
      begin
        S.ACKTimeout := 0;
        Send(S, S.LocalIP, S.RemoteIP, S.LocalPort,
            S.RemotePort, CB_ACK, nil, 0, 0);
      end;
      Inc(i);
    end;
  end;
end;

var
  Fragment: TTCPFragment;

initialization
  LastError := 0;
  SockList := TList.Create;
  FragList := TList.Create;
  NET_AddIdleProc(TCP_Idle);

finalization
  NET_RemoveIdleProc(TCP_Idle);
  { If there's any unclosed socket, close it now }
  while SockList.Count > 0 do
    TCP_Free(SockList[0]);
  { Free the fragments collection }
  while FragList.Count > 0 do
  begin
    Fragment := TTCPFragment(FragList[0]);
    FragList.Remove(Fragment);
    Fragment.Free;
  end;
  FragList.Free;
  { Free the socket collection }
  SockList.Free;
end.
